@@ -1,6 +1,6 @@
 
 /*
-  Kjetil Matheussen, 2005-2012.
+  Kjetil Matheussen, 2005-2013.
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -93,6 +93,7 @@ static float min_buffer_time=-1.0f;
 static float max_buffer_time=40.0f;
 static jack_client_t *client=NULL;
 #define DEFAULT_NUM_CHANNELS 2
+#define DEFAULT_NUM_CHANNELS_SDS 1
 static int num_channels=-1;
 static int bitdepth=0;
 static char *base_filename=NULL;
@@ -103,7 +104,7 @@ static bool show_bufferusage=true;
 static char *meterbridge_type="vu";
 static char *meterbridge_reference="0";
 //static const int vu_len=56;
-#define vu_len 56
+#define vu_len 65
 static int vu_dB=true;
 static float vu_bias=1.0f;
 static int leading_zeros=1;
@@ -141,6 +142,12 @@ static char *hook_cmd_rotate = NULL;
 static char *hook_cmd_timing = NULL;
 static int64_t rotateframe=0;
 
+static bool timemachine_mode = false;
+static bool timemachine_recording = false;
+static float timemachine_prebuffer = 8.0f;
+static int64_t num_frames_written_to_disk = 0;
+static bool program_ended_with_return = false;
+
 /* JACK data */
 static jack_port_t **ports;
 static jack_port_t **ports_meterbridge=NULL;
@@ -150,6 +157,7 @@ static bool jack_has_been_shut_down=false;
 
 static int64_t unreported_overruns=0;
 static int total_overruns=0;
+static int total_xruns=0;
 
 static volatile int freewheel_mode=0;
 
@@ -333,6 +341,9 @@ static int autoincrease_callback(vringbuffer_t *vrb, bool first_call, int readin
     set_high_priority();
     return 0; }
 
+  if(timemachine_mode==true && timemachine_recording==false)
+    return 0;
+
 #if 0
   if(1){
     static jack_transport_state_t prev_state=-1;
@@ -385,14 +396,35 @@ static int findnumports(const char **ports){
 
 static void portnames_add_defaults(void){
   if(cportnames==NULL){
-    cportnames=jack_get_ports(client,NULL,NULL,JackPortIsPhysical|JackPortIsInput);
-    num_cportnames=JC_MAX(DEFAULT_NUM_CHANNELS,findnumports(cportnames));
-    if(num_cportnames==0){
-      fprintf(stderr,"No physical output ports found in your jack setup. Exiting.\n");
-      exit(0);
+
+    {
+      const char **portnames = jack_get_ports(client,NULL,NULL,JackPortIsPhysical|JackPortIsInput);
+      int num_ports    = findnumports(portnames);
+
+      if(num_ports==0){
+        fprintf(stderr,"No physical output ports found in your jack setup. Exiting.\n");
+        exit(0);
+      }
+
+      cportnames = my_calloc(sizeof(char*), num_ports+1);
+      {
+        int i;
+        for(i=0;i<num_ports;i++)
+          cportnames[i] = portnames[i];
+      }
+
+      jack_free(portnames);
     }
-    if(num_channels==-1)
-      num_channels=DEFAULT_NUM_CHANNELS;
+
+    num_cportnames=JC_MAX(DEFAULT_NUM_CHANNELS,findnumports(cportnames));
+    if(num_channels==-1) {
+      if(!strcasecmp("sds",soundfile_format)){
+        num_channels=DEFAULT_NUM_CHANNELS_SDS;
+      }else{
+        num_channels=DEFAULT_NUM_CHANNELS;
+      }
+    }
+
   }else
     if(num_channels==-1)
       num_channels=num_cportnames;
@@ -424,7 +456,7 @@ static void portnames_add(char *name){
     new_outportnames          = jack_get_ports(client,pattern,"",0);
     //char **new_outportnames = (char**)jack_get_ports(client,"system:capture_1$","",0);
     add_ch                    = findnumports(new_outportnames);
-		free(pattern);
+    free(pattern);
   }else{
     new_outportnames          = my_calloc(1,sizeof(char*));
     new_outportnames[0]       = name;
@@ -450,8 +482,7 @@ static void portnames_add(char *name){
   }
 }
 
-
-static const char **portnames_get_connections(int ch){
+static const char **portnames_get_connections(int ch, bool *using_calloc){
   if(ch>=num_cportnames)
     return NULL;
   else{      
@@ -465,9 +496,11 @@ static const char **portnames_get_connections(int ch){
 
     if(jack_port_flags(port) & JackPortIsInput){
       ret    = jack_port_get_all_connections(client,port);
+      *using_calloc = false;
     }else{
       ret    = my_calloc(2,sizeof(char*));
       ret[0] = cportnames[ch];
+      *using_calloc = true;
     }
     
     return ret;
@@ -532,6 +565,8 @@ static void print_console_top(void){
     for(lokke=0;lokke<vu_len;lokke++)
       putchar(c);
     printf("|");print_ln();
+    printf("%c[0m",0x1b); // reset colors
+    fflush(stdout);
   }else{
     //print_ln();
   }
@@ -558,9 +593,12 @@ static void move_cursor_to_top(void){
            : show_bufferusage
              ? 1
              : 0);
-  msleep(3);
+  printf("%c[0m",0x1b); // reset colors
+  fflush(stdout);
 }
 
+static char *vu_not_recording="-----------Press <Return> to start recording------------";
+
 // Console colors:
 // http://www.linuxjournal.com/article/8603
 
@@ -612,22 +650,49 @@ static void print_console(bool move_cursor_to_top_doit,bool force_update){
         vol[1] = '0'+ch;
       }
       
-      for(i=0;i<vu_len;i++)
-        if(vu_peaks[ch]==i && vu_peakvals[ch]>0.0f)
-          vol[4+i] = '*';
-        else if(i<=pos && val>0.0f)
-          vol[4+i] = '-';
-        else
-          vol[4+i] = ' ';
-      
-      if(vu_peakvals[ch]>=1.0f){
-        vol[4+vu_len]='!';
-        printf("%c[31m",0x1b); //red color
-        puts(vol);
+      if (timemachine_mode==true && timemachine_recording==false) {
+
+        for(i=0;i<pos && val>0.0f;i++)
+          vol[4+i] = vu_not_recording[i];
+
+        vol[4+pos]='\0';
+
+        if(vu_peakvals[ch]>=1.0f)
+          printf("%c[31m",0x1b); // Peaking, show red color
+
+        printf(vol);
+
+        for(;i<vu_len;i++)
+          vol[4+i] = vu_not_recording[i];
+
+        printf("%c[33m",0x1b); // Yellow color
+
+        vol[i+4]='\0';
+        printf(vol+4+pos);
+
         printf("%c[36m",0x1b); // back to cyan
-      }else{
-        vol[4+vu_len]='|';
-        puts(vol);
+        printf("|\n");
+
+      } else {
+
+        for(i=0;i<vu_len;i++)
+          if(vu_peaks[ch]==i && vu_peakvals[ch]>0.0f)
+            vol[4+i] = '*';
+          else if(i<=pos && val>0.0f)
+            vol[4+i] = '-';
+          else
+            vol[4+i] = ' ';
+
+        if(vu_peakvals[ch]>=1.0f){
+          vol[4+vu_len]='!';
+          printf("%c[31m",0x1b); //red color
+          puts(vol);
+          printf("%c[36m",0x1b); // back to cyan
+        }else{
+          vol[4+vu_len]='|';
+          puts(vol);
+        }
+
       }
     }
   }
@@ -638,19 +703,34 @@ static void print_console(bool move_cursor_to_top_doit,bool force_update){
     float buflen      = buffers_to_seconds(num_buffers);
     float bufleft     = buffers_to_seconds(num_bufleft);
     int   recorded_seconds = (int)frames_to_seconds(num_frames_recorded);
+    if(timemachine_mode==true)
+      recorded_seconds = (int)frames_to_seconds(num_frames_written_to_disk);
     int   recorded_minutes = recorded_seconds/60;
+
+    char buffer_string[1000];
+    {
+      sprintf(buffer_string,"%.2fs./%.2fs",bufleft,buflen);
+      int len_buffer=strlen(buffer_string);
+      int i;
+      for(i=len_buffer;i<14;i++)
+        buffer_string[i]=' ';
+      buffer_string[i]='\0';
+    }
+
     printf("%c[32m"
-           "Buffer: %.2fs./%.2fs. "
+           "Buffer: %s"
            "  Time: %d.%s%dm.  %s"
            "DHP: [%c]  "
-           "Overruns: %d"
+           "Overruns: %d  "
+           "Xruns: %d"
            "%c[0m",
            //fmaxf(0.0f,buflen-bufleft),buflen,
            0x1b, // green color
-           bufleft,buflen,
+           buffer_string,
            recorded_minutes, recorded_seconds%60<10?"0":"", recorded_seconds%60, recorded_minutes<10?" ":"", 
            disk_thread_has_high_priority?'x':' ',
            total_overruns,
+           total_xruns,
            0x1b // reset color
            );
     print_ln();
@@ -658,6 +738,8 @@ static void print_console(bool move_cursor_to_top_doit,bool force_update){
     printf("%c[0m",0x1b); // reset colors
     fprintf(stderr,"%c[0m",0x1b); // reset colors
   }
+  fflush(stdout);
+  fflush(stderr);
 }
 
 
@@ -750,8 +832,8 @@ static void print_message(const char *fmt, ...){
     va_start(argp,fmt);
     fprintf(stderr,"%c[31m" MESSAGE_PREFIX,0x1b);   // set red color
     vfprintf(stderr,fmt,argp);
-    //fprintf(stderr,"%c[0m",0x1b); // reset colors
-    //msleep(1);
+    fprintf(stderr,"%c[0m",0x1b); // reset colors
+    fflush(stderr);
     va_end(argp);
   }else{
 
@@ -969,6 +1051,8 @@ static int64_t overruns=0;
 #if HAVE_LIBLO
 bool queued_file_rotate=false;
 void osc_stop() { sem_post(&stop_sem); }
+void osc_tm_start() { timemachine_recording=true; }
+void osc_tm_stop() { program_ended_with_return=true; osc_stop(); }
 #endif
 
 static struct timespec rtime;
@@ -1089,7 +1173,7 @@ static int open_soundfile(void){
   case 32: subformat = SF_FORMAT_PCM_32;
     break;
   default:
-    if(!strcasecmp("flac",soundfile_format)){
+    if(!strcasecmp("flac",soundfile_format) || !strcasecmp("sds",soundfile_format)){
       bitdepth=24;
       subformat=SF_FORMAT_PCM_24;
 #if HAVE_OGG
@@ -1144,7 +1228,7 @@ static void close_soundfile(void){
 #endif
   }
 
-  hook_file_closed(filename, total_overruns, disk_errors);
+  hook_file_closed(filename, total_overruns + total_xruns, disk_errors);
 
   if (overruns > 0) {
     print_message("jack_capture failed with a total of %d overruns.\n", total_overruns);
@@ -1152,9 +1236,15 @@ static void close_soundfile(void){
   }
   if (disk_errors > 0)
     print_message("jack_capture failed with a total of %d disk errors.\n",disk_errors);
+  if (total_xruns > 0)
+    print_message("jack_capture encountered %d jack x-runs.\n", total_xruns);
+
+  disk_errors = 0;
+  total_overruns = 0;
+  total_xruns = 0;
 }
 
-static int rotate_file(size_t frames){
+static int rotate_file(size_t frames, int reset_totals){
 	store_sync=0;
 	// Explanation: new file will already contain the CURRENT buffer!
 	// but sync-timeframe will only be saved on the start of next jack cycle.
@@ -1170,13 +1260,29 @@ static int rotate_file(size_t frames){
   print_message("Closing %s, and continue writing to %s.\n",filename,filename_new);
   num_files++;
 
-  hook_file_rotated(filename, filename_new, num_files, total_overruns, disk_errors);
+  hook_file_rotated(filename, filename_new, num_files, total_overruns + total_xruns, disk_errors);
 
   free(filename);
   filename=filename_new;
   disksize=0;
-  if(!open_soundfile()) return 0;
 
+  if (reset_totals) {
+    /* reset totals on file-rotate */
+    disk_errors = 0;
+    total_overruns = 0;
+    total_xruns = 0;
+
+    if (overruns > 0) {
+      print_message("jack_capture failed with a total of %d overruns.\n", total_overruns);
+      print_message("   try a bigger buffer than -B %f\n",min_buffer_time);
+    }
+    if (disk_errors > 0)
+      print_message("jack_capture failed with a total of %d disk errors.\n",disk_errors);
+    if (total_xruns > 0)
+      print_message("jack_capture encountered %d jack x-runs.\n", total_xruns);
+  }
+
+  if(!open_soundfile()) return 0;
 
   return 1;
 }
@@ -1190,17 +1296,17 @@ static int handle_filelimit(size_t frames){
 
   if(is_using_wav && (disksize + ((int64_t)new_bytes) >= UINT32_MAX-(1024*1024))){ // (1024*1024) should be enough for the header.
     print_message("Warning. 4GB limit on wav file almost reached.");
-    if (!rotate_file(frames)) return 0;
+    if (!rotate_file(frames, false)) return 0;
   }
 #if HAVE_LIBLO
   else if (queued_file_rotate) {
     queued_file_rotate=false;
     print_message("Note. file-name rotation request received.");
-    if (!rotate_file(frames)) return 0;
+    if (!rotate_file(frames, true)) return 0;
   }
 #endif
   else if (rotateframe > 0 && disksize > (rotateframe*bytes_per_frame*num_channels) ) {
-    if (!rotate_file(frames)) return 0;
+    if (!rotate_file(frames, false)) return 0;
 	}
   disksize+=new_bytes;
   return 1; }
@@ -1259,6 +1365,8 @@ static int mp3_write(void *das_data,size_t frames){
 
 static int disk_write(void *data,size_t frames){
 
+  num_frames_written_to_disk += frames;
+
   if(write_to_stdout==true)
     return stdout_write(data,frames);
 
@@ -1307,25 +1415,25 @@ static int disk_write_overruns(int num_overruns){
 }
 
 static void disk_thread_control_priority(void){
+  int adjusted_writing_size = vringbuffer_writing_size(vringbuffer);
+  if (timemachine_mode==true)
+    adjusted_writing_size += seconds_to_blocks(timemachine_prebuffer*jack_samplerate);
+
   if(1
      && disk_thread_has_high_priority==false
-     && vringbuffer_reading_size(vringbuffer) >= vringbuffer_writing_size(vringbuffer)
+     && vringbuffer_reading_size(vringbuffer) >= adjusted_writing_size
      && use_jack_freewheel==false
      )
     {
       if(set_high_priority()==true){
         disk_thread_has_high_priority=true;
-        print_message("Less than half of the buffer used. Setting higher priority for the disk thread.\n");
+        print_message("Less than half the buffer used. Setting higher priority for the disk thread.\n");
       }else{
         static bool message_sent=false;
         if(message_sent==false)
           print_message("Error. Could not set higher priority for disk thread.\n");
         message_sent=true; } } }
 
-static bool timemachine_mode = false;
-static bool timemachine_recording = false;
-static float timemachine_prebuffer = 6.0f;
-
 static enum vringbuffer_receiver_callback_return_t disk_callback(vringbuffer_t *vrb,bool first_time,void *element){
   static bool printed_receive_message=false;
   buffer_t *buffer=(buffer_t*)element;
@@ -1342,6 +1450,10 @@ static enum vringbuffer_receiver_callback_return_t disk_callback(vringbuffer_t *
       return VRB_CALLBACK_DIDNT_USE_BUFFER;
   }
 
+  if (timemachine_mode==true && printed_receive_message==false){
+    print_message("Recording. Press <Return> to stop.\n");
+    printed_receive_message=true; }
+
   if(use_jack_transport==true && printed_receive_message==false){
     print_message("Received JackTranportRolling. Recording.\n");
     printed_receive_message=true; }
@@ -1591,9 +1703,10 @@ static int process(jack_nframes_t nframes, void *arg){
   return 0;
 }
 
-
-
-
+static int xrun(void *arg){
+	total_xruns++;
+	return 0;
+}
 
 
 /////////////////////////////////////////////////////////////////////
@@ -1700,16 +1813,25 @@ static int init_meterbridge_ports(){
 /////////////////////////////////////////////////////////////////////
 
 
+static void free_jack_connections(bool using_calloc, const char **connections) {
+  if (connections==NULL)
+    return;
+  if (using_calloc)
+    free(connections);
+  else
+    jack_free(connections);
+}
+
 static int compare(const void *a, const void *b){
   return strcmp((const char*)a,(const char*)b);
 }
 
-
 static int reconnect_ports_questionmark(void){
   int ch;
 
   for(ch=0;ch<num_channels;ch++){
-    const char **connections1 = portnames_get_connections(ch);
+    bool using_calloc;
+    const char **connections1 = portnames_get_connections(ch, &using_calloc);
     const char **connections2 = jack_port_get_all_connections(client,ports[ch]);
 
     int memb1 = findnumports(connections1);
@@ -1719,8 +1841,8 @@ static int reconnect_ports_questionmark(void){
       continue;
       
     if(memb1!=memb2){
-      free(connections1);
-      free(connections2);
+      free_jack_connections(using_calloc, connections1);
+      free_jack_connections(false, connections2);
       return 1;
     }
 
@@ -1732,15 +1854,15 @@ static int reconnect_ports_questionmark(void){
       for(lokke=0;lokke<memb1;lokke++){
         //printf("connect_ports \"%s\" \"%s\" \n",connections1[lokke],connections2[lokke]);
         if(strcmp(connections1[lokke],connections2[lokke])){
-          free(connections1);
-          free(connections2);
+          free_jack_connections(using_calloc, connections1);
+          free_jack_connections(false, connections2);
           return 1;
         }
       }
     }
-      
-    free(connections1);
-    free(connections2);
+
+    free_jack_connections(using_calloc, connections1);
+    free_jack_connections(false, connections2);
   }
 
   return 0;
@@ -1761,7 +1883,7 @@ static void disconnect_ports(jack_port_t** ports){
       for(;connections[lokke]!=NULL;lokke++)
         jack_disconnect(client,connections[lokke],jack_port_name(ports[ch]));
     
-    free(connections);
+    free_jack_connections(false, connections);
   }
 }
 
@@ -1775,16 +1897,17 @@ static void connect_ports(jack_port_t** ports){
   for(ch=0;ch<num_channels;ch++){
     int lokke = 0;
 
-    const char **connections = portnames_get_connections(ch);
+    bool using_calloc;
+    const char **connections = portnames_get_connections(ch, &using_calloc);
 
     while(connections && connections[lokke] != NULL){
       int err=jack_connect(client,connections[lokke],jack_port_name(ports[ch]));
       if(err!=0)
-	print_message("\nCan not connect input port %s to %s, errorcode %s\n",
-		    jack_port_name (ports[ch]), connections[lokke],strerror(err));      
+	print_message("\nCould not connect input port %s to %s, errorcode %s\n",
+                      jack_port_name (ports[ch]), connections[lokke],strerror(err));      
       lokke++;
     }
-    free(connections);
+    free_jack_connections(using_calloc, connections);
   }
 }
 
@@ -1937,9 +2060,25 @@ static void start_jack(void){
 static pthread_t keypress_thread={0};
 static void* keypress_func(void* arg){
   char gakk[64];
+
   turn_off_echo();
-  if(fgets(gakk,49,stdin) != NULL)
+
+  char *fgets_result;
+
+ again:
+
+  fgets_result = fgets(gakk,49,stdin);
+
+  if(timemachine_mode==true && timemachine_recording==false){
+    timemachine_recording = true;
+    goto again;
+  }
+
+  if(fgets_result!=NULL)
     ungetc('\n',stdin);
+
+  program_ended_with_return = true;
+
   sem_post(&stop_sem);
   return NULL;
 }
@@ -2018,6 +2157,9 @@ static const char *advanced_help =
   "[--hook-close c] or [-Hc c]      -> command to execute when closing the session. (see below)\n"
   "[--hook-rotate c] or [-Hr c]     -> command to execute on file-name-rotation. (see below)\n"
   "[--hook-timing c] or [-Ht c]     -> callback when first audio frame is received. (see below)\n"
+  "[--timemachine] or [-tm]         -> jack_capture operates in \"timemachine\" mode.\n"
+  "[--timemachine-prebuffer s]      -> Specify (in seconds) how long time to prebuffer in timemachine mode.\n"
+  "[ -tmpb s]                       -> ------------------------ \"\" ----------------------------------------\n"
   "\n"
   " All hook options take a full-path to an executable as argument.\n"
   " The commands are executed in a fire-and-forget style upon internal events.\n"
@@ -2070,6 +2212,8 @@ static const char *osc_help =
   "\n"
   "  /jack_capture/stop        (no arguments) -- stop recording and exit\n"
   "  /jack_capture/rotate      (no arguments) -- rotate file-name\n"
+  "  /jack_capture/tm/start    (no arguments) -- in timemachine-mode: start recording.\n"
+  "  /jack_capture/tm/stop     (no arguments) -- in timemachine-mode: stop recording.\n"
   "\n"
   "Example:\n"
   "  jack_capture -O 7777\n"
@@ -2158,6 +2302,8 @@ void init_arguments(int argc, char *argv[]){
       OPTARG("--hook-timing","-Ht") hook_cmd_timing = OPTARG_GETSTRING();
       OPTARG("--timestamp","-S") create_tme_file=true;
       OPTARG("--rotatefile","-Rf") rotateframe = OPTARG_GETINT();
+      OPTARG("--timemachine","-tm") timemachine_mode = true;
+      OPTARG("--timemachine-prebuffer","-tmpb") timemachine_prebuffer=OPTARG_GETFLOAT();
       OPTARG_LAST() base_filename=OPTARG_GETSTRING();
     }OPTARGS_END;
 
@@ -2181,6 +2327,11 @@ void init_arguments(int argc, char *argv[]){
       min_buffer_time = DEFAULT_MIN_BUFFER_TIME;
   }
 
+  if(timemachine_mode==true) {
+    min_buffer_time += timemachine_prebuffer;
+    max_buffer_time += timemachine_prebuffer;
+  }
+
   verbose_print("main() find default file format\n");
   if(soundfile_format_is_set==false){
     if(num_channels>2)
@@ -2341,6 +2492,7 @@ void init_various(void){
   // Init jack 2
   {
     jack_set_process_callback(client, process, NULL);
+    jack_set_xrun_callback(client, xrun, NULL);
 
     jack_on_shutdown(client, jack_shutdown, NULL);
 
@@ -2390,9 +2542,14 @@ void init_various(void){
                     base_filename,
                     recording_time);  
     }else{
-      if(silent==false)
-        print_message("Recording to \"%s\". Press <Return> or <Ctrl-C> to stop.\n",base_filename);        
-      //fprintf(stderr,"Recording to \"%s\". Press <Return> or <Ctrl-C> to stop.\n",base_filename);        
+      if(silent==false) {
+        if (timemachine_mode==true) {
+          print_message("Waiting to start recording of \"%s\"\n",base_filename);
+          print_message("Press <Ctrl-C> to stop recording and quit.\n");
+        }else
+          print_message("Recording to \"%s\". Press <Return> or <Ctrl-C> to stop.\n",base_filename);
+        //fprintf(stderr,"Recording to \"%s\". Press <Return> or <Ctrl-C> to stop.\n",base_filename);        
+      }
     }
   }
 
@@ -2415,7 +2572,9 @@ void wait_until_recording_finished(void){
   if(use_jack_freewheel==true)
     print_message("Waiting for Jack Freewheeling .\n");
   
-  sem_wait(&stop_sem);
+  while(sem_wait(&stop_sem)==-1)
+    print_message("Warning: sem_wait failed: %s",strerror(errno));
+
   turn_on_echo();
   if(helper_thread_running==1){
     //      if(use_vu || show_bufferusage)
@@ -2494,6 +2653,8 @@ int main (int argc, char *argv[]){
   //get_free_mem();
   //mainpid=getpid();
 
+  char **org_argv = argv;
+
   // remove exe name from argument list.
   argv = &argv[1];
   argc = argc-1;
@@ -2519,6 +2680,12 @@ int main (int argc, char *argv[]){
 
   stop_recording_and_cleanup();
 
+  if (timemachine_mode==true && program_ended_with_return==true){
+    execv (org_argv[0], (char *const *) org_argv);
+    print_message("Error: exec returned.\n");
+    exit(127);
+  }
+
   return 0;
 }
 
